import java.util.*;
import java.io.*;

class PumpVariable {
    double beta;
    double lambda[];
    
    PumpVariable(int n) {
	lambda = new double[n];
    }
    
    double norm_lambda() {
	double norm = 0;
	for(int i = 0; i < lambda.length; i++) {
	    norm += lambda[i];
	}
	return norm;
    }

    public String toString() {
	StringBuffer b = new StringBuffer(80);
	b.append(Math.floor(beta*1000)/1000);
	for(int i = 0; i < lambda.length; i++) {
	    b.append(" ").append(Math.floor(lambda[i]*1000)/1000);
	}
	return b.toString();
    }
}



public class PumpSimulator {

    // utility class
    class Interval {
	double a,b;
	Interval(double left, double right) {
	    a = left;
	    b = right;
	}
    }

    // here's the data
    static double t[] = 
    { 94.320, 15.720, 62.880, 125.760, 5.240, 31.440, 1.048, 1.048, 2.096, 10.480};
    static double s[] = 
    { 5, 1, 5, 14, 3, 19, 1, 1, 4, 22};
    

    // data
    double alpha, gamma, delta, maxt;
    int num_pumps;
    // sweep
    double psi[], xi;
    PumpVariable chain_next, B;
    // catalysts
    double  norm_lambda_star, nls[], xis[];
    PumpVariable Y, catalysts[], all_chains[];
    // public variables used for verbosity
    public double L;
    public boolean coalesced, fully_coupled;
    // my own random number generator and other stuff
    MyRandom r;
    Interval C;
    Vector J;


    // constructor
    public PumpSimulator(double aT[], double anS[], double anAlpha, 
			 double aGamma, double aDelta, int numSweeps) {
	// copy the parameters
	t = aT;
	s = anS;
	alpha = anAlpha;
	gamma = aGamma;
	delta = aDelta;
	num_pumps = t.length;

	// set up some useful objects
	r = new MyRandom();
	J = new Vector();

	// create some extra variables
	B = new PumpVariable(num_pumps);
	chain_next = new PumpVariable(num_pumps);
	psi = new double[num_pumps+1]; // for Gibbs sweeps
	maxt = 0; // useful to compute the superset J of K_B
	for(int i=0; i < num_pumps; i++) {
	    maxt = ( maxt < t[i] ) ? t[i] : maxt;
	}

    }


    public PumpVariable bigSweep(PumpVariable ref_chain, int num_sweeps,
				 int cats_per_sweep) {

	// FIRST, we make the independence sampler step	
	B.beta = r.nextGamma(gamma)/delta;
	for(int i=0; i < num_pumps; i++) {
	    B.lambda[i] = r.nextGamma(alpha)/ B.beta;
	}

	// propagate the reference chain
	// by seeing if it is in K_B or not
	L = - Math.log(r.nextDouble()); // the log of psi
	double test = 0;
	for(int j=0; j < num_pumps; j++) {
	    L += s[j]*Math.log(B.lambda[j]) - t[j]*B.lambda[j];
	    test += s[j]*ref_chain.lambda[j] - t[j]*ref_chain.lambda[j];
	}

	if( test >= L ) { // we're in K_B
	    chain_next.beta = ref_chain.beta;
	    System.arraycopy(ref_chain.lambda,0,chain_next.lambda,0,num_pumps);
	} else { // we're not
	    chain_next.beta = B.beta;
	    System.arraycopy(B.lambda,0,chain_next.lambda,0,num_pumps);
	}
	// Since K_B  ={\lambda : <s,log\lambda> - <t,\lambda> >= L}
	// we must have K_B \subseteq {\norm{\lambda} \in [0, -L/t_max]}. 
	// We shall use this bounding interval in the algorithm below 
	L = -L / maxt;

	// NEXT, we compute the reference chains in J
	// by producing an update map which is hopefully finitely coupled
	// over the interval [0, -L/t_max]. 
	
	if( !J.isEmpty() ) 
	    J.removeAllElements(); // J = \emptyset
	// generate the basic update and the accompanying catalysts
	initCoupledGibbsSweep(cats_per_sweep);

	// now calculate the coverage to see if we are finitely coupled
	for(int i = 0; i < cats_per_sweep; i++) {
	    double a = catalysts[i].beta;
	    double b = Math.log(xis[i]) - catalysts[i].beta * 
		(nls[i] - delta);
	    double c = delta * (Math.log(xis[i]) - catalysts[i].beta *
				nls[i]) - psi[0];
	    C = new Interval( (-b - Math.sqrt(b*b - 4*a*c))/(2*a),
			      (-b + Math.sqrt(b*b - 4*a*c))/(2*a) );
	    addToJ(C);
//  	    System.out.println("added ["+
//  			       (-b - Math.sqrt(b*b - 4*a*c))/(2*a*L)
//  			       +", " + 
//  			       (-b + Math.sqrt(b*b - 4*a*c))/(2*a*L)+
//  			       "] lambda_star = " + nls[i] +
//  			       "full="+isFullJ());
	}

	if( Double.isNaN(L) || !isFullJ() ) { 
	    // L is infinite so we can't coalesce
	    coalesced = false;
	    fully_coupled = false;
	    // propagate the reference chain
	    chain_next = evalCoupledGibbsSweep(chain_next);
	    return chain_next; // done
	} else { 
	    // we have a fully coupled map on [0,L]
	    fully_coupled = true; // but do we coalesce?
	    // create a list of all the chains we must track
	    all_chains = new PumpVariable[cats_per_sweep+1];
	    all_chains[0] = evalCoupledGibbsSweep(chain_next);
	    System.arraycopy(catalysts, 0, all_chains, 1, cats_per_sweep);

	    // Now let's update all reference chains by performing m 
	    // extra Gibbs sweeps with k catalyst proposals at each sweep
	    for( int sweep = 1; sweep < num_sweeps; sweep++) {
		initCoupledGibbsSweep(cats_per_sweep);
		for( int c = 0; c < all_chains.length; c++) {
		    all_chains[c] = evalCoupledGibbsSweep(all_chains[c]);
		}
	    }
	
	    // now see if we had coalescence
	    coalesced = true;
	    for( int c1 = 1; c1 < all_chains.length; c1++) {
		if( all_chains[c1].beta != all_chains[0].beta )
		    coalesced = false;
		for(int l = 0; l < num_pumps; l++)
		    if( all_chains[c1].lambda[l] != all_chains[0].lambda[l] )
			coalesced = false;
		if( !coalesced )
		    break;
	    }
	    return all_chains[0]; // this is the propagated  value of refChain
	}
    }

    
    // this adds another interval to the list J which represents 
    // their union
    void addToJ(Interval I) {
	int numIntervals = J.size();
	boolean done = false;
	int i = 0;
	while( i < numIntervals && !done) {
	    Interval K = (Interval)J.elementAt(i);
	    if( K.a > I.a) {
		J.insertElementAt(I,i);
		done = true;
	    }
	    i++;
	}
	if( !done )
	    J.addElement(I);
    }


    boolean isFullJ() {
	int numIntervals = J.size();
	boolean gap = false, stop = false;
	int i = 0;
	double x = 0.0;
	while( (i < numIntervals) && !gap && !stop) {
	    Interval I = (Interval)J.elementAt(i);
	    if( I.a > x )
		gap = true;
	    else 
		x = I.b;	    
	    if( x > L )
		stop = true;
	    i++;
	}
	return (x > L);
    }

    void initCoupledGibbsSweep(int num_catalysts) {
	// begin by generating the random variables needed for
	// and ordinary Gibbs sweep
	psi[0] = r.nextGamma(gamma + num_pumps*alpha);
	for(int i = 0; i < num_pumps; i++) {
	    psi[i+1] = r.nextGamma(alpha + s[i]);
	}
	// next we generate all the catalysts
	catalysts = new PumpVariable[num_catalysts];
	xis = new double[num_catalysts];
	nls = new double[num_catalysts];

	// next we build the corresponding catalysts
	for(int i = 0; i < num_catalysts; i++) {
	    nls[i] = r.nextDouble() * L; // norm lambda star
	    catalysts[i] = new PumpVariable(num_pumps);
	    catalysts[i].beta = r.nextGamma(gamma + num_pumps*alpha)/(delta + nls[i]);
	    for(int k = 0; k < num_pumps; k++) {
		catalysts[i].lambda[k] = r.nextGamma(alpha + s[k])/
		    (catalysts[i].beta + t[k]);
	    }
	    xis[i] = r.nextDouble(); // and the test uniforms
	}	
    }

    PumpVariable evalCoupledGibbsSweep(PumpVariable from) {
	PumpVariable to = new PumpVariable(num_pumps);

	// first evaluate the ordinary Gibbs sweep
	to.beta = psi[0]/(delta + from.norm_lambda());
	for(int i = 0; i < num_pumps; i++) {
	    to.lambda[i] = psi[i+1]/(to.beta + t[i]);
	}
	// next cycle through the catalysts
	for(int j = 0; j < catalysts.length; j++) {
	    if( Math.exp((to.beta - catalysts[j].beta) *
			 (from.norm_lambda() - nls[j])) > xis[j] ) {
		to.beta = catalysts[j].beta;
		System.arraycopy(catalysts[j].lambda,0,to.lambda,0,num_pumps);
	    }				
	}
	return to;
    }

    static int number_of_sweeps = 20;
    static int number_of_catalysts_per_sweep = 20;
    static int number_of_iterations = 50;
    static boolean verbose = false;
    static boolean force_output = false;


    public static void main(String args[]) throws IOException {

	// read in the command line arguments
	for( int i = 0; i < args.length; i++) {
	    if( args[i].charAt(0) != '-' ) {
		System.out.println("Unrecognized command Line switch");
	    }
	    switch( args[i].charAt(1) ) {
	    case 'v':
	    case 'V': 
		verbose = true;
		break;
	    case 's':
	    case 'S':
		try {
		    number_of_sweeps = Integer.parseInt(args[++i]);
		} catch (NumberFormatException e){
		   System.out.println("Unrecognized command Line switch");
		   number_of_sweeps = 35;
		}
		break;
	    case 'c':
	    case 'C':
		try {
		    number_of_catalysts_per_sweep = Integer.parseInt(args[++i]);
		} catch (NumberFormatException e){
		   System.out.println("Unrecognized command Line switch");
		   number_of_catalysts_per_sweep = 10;
		}
		break;
	    case 'i':
	    case 'I':
		try {
		    number_of_iterations = Integer.parseInt(args[++i]);
		} catch (NumberFormatException e){
		   System.out.println("Unrecognized command Line switch");
		   number_of_iterations = 50;
		}
		break;
	    case 'f':
	    case 'F': // force everything to be written in pump.dat
		force_output = true;
	    }
	}


	System.out.println("PumpSimulator is running");
	
	// create a new pump simulator
	PumpSimulator ps = new PumpSimulator(t,s,1.802,0.01,1, 10);


	// initialize the reference chain in some way
	PumpVariable ref_chain = new PumpVariable(10);
	ref_chain.beta = 1;
	for(int j=0; j < 10; j++) {
	    ref_chain.lambda[j] = 1; // say
	}

	if( verbose ) {
	    System.out.println("Number of iterations: " + number_of_iterations
			       + " using " + number_of_sweeps 
			       + " sweeps/iteration and " +
			       number_of_catalysts_per_sweep +
			       " catalysts/sweep.");
	    System.out.println("X_0 = ["+ref_chain+"]");
	}

	// open the output file
	FileWriter fw = new FileWriter("pump.dat");
	BufferedWriter f = new BufferedWriter(fw, 8192);
	f.write("#h beta lambda1 lambda2 lambda3 lambda4 lambda5 lambda6 lambda7 lambda8 lambda9 lambda10");
	if( force_output )
	    f.write(" forced output NOT IID!\n");
	else
	    f.write("\n"); 

	boolean first = true;
	int j = 1;
	PumpVariable old_chain = new PumpVariable(10);
	for(int i = 0; i < number_of_iterations; i++) {
	    // save the state just prior to testing for coalescence
	    old_chain.beta = ref_chain.beta;
	    System.arraycopy(ref_chain.lambda,0,old_chain.lambda,0,10);
	    // do the coalescence thing
	    ref_chain = ps.bigSweep(ref_chain, number_of_sweeps, 
				    number_of_catalysts_per_sweep);
	    if( verbose ) {
		System.out.print("" + (i+1) + ": K_B \\subset [0, "+ps.L+"]");
		if( ps.fully_coupled )
		    System.out.print(" - fully coupled");
		else
		    System.out.print(" - not coupled");
		if( ps.coalesced ) {
		    System.out.println(" + coalesced");
		    System.out.print(" T_"+j+" = "+i);
		} else
		    System.out.println(" - not coalesced");

		System.out.println(" X_"+(i+1)+" = ["+ref_chain+"]");
	    }

	    if(force_output ) {
		f.write(old_chain.toString());
		f.write("\n");
		if( ps.coalesced )
		    j++;
	    } else if( ps.coalesced ) { 
		j++;
		if( first ) {
		    first = false;
		} else { // write the output to the file
		    f.write(old_chain.toString());
		    f.write("\n");
		}
	    }
	}

	System.out.println("Produced an IID sample of size " + (j-2) );
	// close the file
	f.close();
    }
    
}














